#import "config.typ"

#let is-mat(mat) = {
  if type(mat) != dictionary {return false}
  if mat.at("type", default: none) != "matrix" {return false}
  return true
}

#let _check(mat) = {
  if not is-mat(mat) {
    panic("Argument is not a matrix")
  }
}

#let size(mat) = {
  _check(mat)
  return str(mat.h) + "x" + str(mat.w)
}

#let mat(..args) = {
  let rows = args.pos()
  //rows = rows.map(r => if r == none {()} else {r})
  let h = rows.len()
  let w = if h == 0 {0} else {rows.first().len()}
  if rows.find(r => r.len() != w) != none {
    panic("All rows should have the same length")
  }
  return (
    type: "matrix",
    rows: rows,
    w: w,
    h: h
  )
}

#let copy(original) = {
  _check(original)
  return mat(..original.rows)
}

#let of-size(h, w) = {
  let rows = ((0,) * w,) * h
  return mat(..rows)
}

#let display(mat, ..args) = {
  _check(mat)
  math.mat(..mat.rows, ..args.named())
}

#let add(mat1, mat2) = {
  _check(mat1)
  _check(mat2)
  if mat1.w != mat2.w or mat1.h != mat2.h {
    panic("Can't add matrices of size " + size(mat1) + " and " + size(mat2))
  }

  let mat3 = of-size(mat1.h, mat1.w)

  for y in range(mat1.h) {
    for x in range(mat1.w) {
      mat3.rows.at(y).at(x) = mat1.rows.at(y).at(x) + mat2.rows.at(y).at(x)
    }
  }

  return mat3
}

#let sub(mat1, mat2) = {
  _check(mat1)
  _check(mat2)
  if mat1.w != mat2.w or mat1.h != mat2.h {
    panic("Can't subtract matrices of size " + size(mat1) + " and " + size(mat2))
  }

  let mat3 = of-size(mat1.h, mat1.w)

  for y in range(mat1.h) {
    for x in range(mat1.w) {
      mat3.rows.at(y).at(x) = mat1.rows.at(y).at(x) - mat2.rows.at(y).at(x)
    }
  }

  return mat3
}

#let mul(mat1, mat2) = {
  _check(mat1)
  _check(mat2)
  if mat1.w != mat2.h {
    panic("Can't multiply matrices of size " + size(mat1) + " and " + size(mat2))
  }

  let mat3 = of-size(mat1.h, mat2.w)

  for y in range(mat1.h) {
    for x in range(mat2.w) {
      let sum = 0
      for i in range(mat1.w) {
        sum += mat1.rows.at(y).at(i) * mat2.rows.at(i).at(x)
      }
      mat3.rows.at(y).at(x) = sum
    }
  }

  return mat3
}

#let transpose(mat) = {
  _check(mat)
  let mat-t = of-size(mat.w, mat.h)

  for y in range(mat.h) {
    for x in range(mat.w) {
      mat-t.rows.at(x).at(y) = mat.rows.at(y).at(x)
    }
  }

  return mat-t
}

#let to-vec(mat) = {
  _check(mat)
  import "vec.typ" as _vec
  return _vec.vec(..mat.rows.map(r => r.first()))
}

#let mul-vec(mat1, vec) = {
  _check(mat1)
  import "vec.typ" as _vec

  if mat1.w != vec.size {
    panic("Can't multiply matrix of size " + size(mat1) + " and vector of size " + _vec.size(vec))
  }

  let mat2 = _vec.to-mat(vec)
  let res = mul(mat1, mat2)

  return to-vec(res)
}

#let mul-row(mat1, row, f) = {
  _check(mat1)
  let mat2 = copy(mat1)

  for x in range(mat1.w) {
    mat2.rows.at(row).at(x) *= f
  }

  return mat2
}

#let div-row(mat1, row, f) = {
  _check(mat1)
  if f == 0 {
    panic("Can't divide by 0")
  }

  return mul-row(mat1, row, 1/f)
}

#let add-row(mat1, row1, row2, f: 1) = {
  _check(mat1)
  let mat2 = copy(mat1)

  for x in range(mat1.w) {
    mat2.rows.at(row1).at(x) += mat1.rows.at(row2).at(x) * f
  }

  return mat2
}

#let round(mat) = {
  _check(mat)
  let mat2 = copy(mat)

  for y in range(mat2.h) {
    for x in range(mat2.w) {
      mat2.rows.at(y).at(x) = calc.round(
        mat2.rows.at(y).at(x),
        digits: config.PRECISION
      )
    }
  }

  return mat2
}

#let co-mat(mat1, elim-x, elim-y) = {
  _check(mat1)
  let mat2 = of-size(mat1.h - 1, mat1.w - 1)
  for y in range(mat2.h) {
    for x in range(mat2.w) {
      let ox = if x < elim-x {x} else {x + 1}
      let oy = if y < elim-y {y} else {y + 1}
      mat2.rows.at(y).at(x) = mat1.rows.at(oy).at(ox)
    }
  }
  return mat2
}

#let det(mat) = {
  _check(mat)
  if mat.w != mat.h {
    panic("Matrix must be square")
  }

  if mat.w == 0 {
    return 1
  }

  let res = 0
  for y in range(mat.h) {
    let sign = (1 - 2 * calc.rem(y, 2))
    let co-fact = det(co-mat(mat, 0, y))
    res += sign * mat.rows.at(y).at(0) * co-fact
  }
  return res
}

#let det-steps(mat) = {
  _check(mat)
  if mat.w != mat.h {
    panic("Matrix must be square")
  }

  let res = []
  let lines = ()
  res += display(mat, delim: "|")
  let ap = $&$.body

  let terms = ((1, mat,),)
  for _ in range(mat.w - 2) {
    let line = [
      #ap =
    ]
    
    for (i, term) in terms.enumerate() {
      let (coef, m) = term

      let sub-terms = ()
      for y in range(m.h) {
        let c = (1 - 2 * calc.rem(y, 2)) * coef * m.rows.at(y).at(0)
        let m2 = co-mat(m, 0, y)
        if c != 0 {
          sub-terms.push((c, m2))
        }

        if y + i != 0 and c >= 0 {
          line += [+]
        }
        if c != 1 {
          if c == -1 {
            line += [-]
          } else {
            line += [#c #sym.dot]
          }
        }
        line += display(m2, delim: "|")
      }

      terms.at(i) = sub-terms
    }
    lines.push(line + linebreak())
    terms = terms.join()
  }

  res += lines.join()

  res += [#ap =]
  let total = 0
  for (i, term) in terms.enumerate() {
    let (coef, m) = term
    let value = coef * det(m)
    total += value
    if value != 0 {
      if i != 0 and value >= 0 {
        res += [+]
      }
      res += [#value]
    }
  }
  res += [=#total]
  return math.equation(res)
}